// Created by: NW,JPB,CAL
// Copyright (c) 1991-1999 Matra Datavision
// Copyright (c) 1999-2014 OPEN CASCADE SAS
//
// This file is part of Open CASCADE Technology software library.
//
// This library is free software; you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License version 2.1 as published
// by the Free Software Foundation, with special exception defined in the file
// OCCT_LGPL_EXCEPTION.txt. Consult the file LICENSE_LGPL_21.txt included in OCCT
// distribution for complete text of the license and disclaimer of any warranty.
//
// Alternatively, this file may be used under the terms of Open CASCADE
// commercial license or contractual agreement.

#ifndef _Graphic3d_AspectLine3d_HeaderFile
#define _Graphic3d_AspectLine3d_HeaderFile

#include <Aspect_AspectLineDefinitionError.hxx>
#include <Aspect_TypeOfLine.hxx>
#include <Graphic3d_ShaderProgram.hxx>
#include <Standard.hxx>
#include <Standard_Real.hxx>
#include <Standard_Type.hxx>
#include <Quantity_ColorRGBA.hxx>

//! Creates and updates a group of attributes for 3d line primitives.
//! This group contains the color, the type of line, and its thickness.
class Graphic3d_AspectLine3d : public Standard_Transient
{
  DEFINE_STANDARD_RTTIEXT(Graphic3d_AspectLine3d, Standard_Transient)
public:

  //! Creates a context table for line primitives
  //! defined with the following default values:
  //!
  //! Color = Quantity_NOC_YELLOW;
  //! Type  = Aspect_TOL_SOLID;
  //! Width = 1.0;
  Standard_EXPORT Graphic3d_AspectLine3d();
  
  //! Creates a context table for line primitives defined with the specified values.
  //! Warning: theWidth is the "line width scale factor".
  //! The nominal line width is 1 pixel.
  //! The width of the line is determined by applying the line width scale factor to this nominal line width.
  //! The supported line widths vary by 1-pixel units.
  Standard_EXPORT Graphic3d_AspectLine3d (const Quantity_Color&   theColor,
                                          const Aspect_TypeOfLine theType,
                                          const Standard_Real     theWidth);

  //! Return color.
  const Quantity_ColorRGBA& ColorRGBA() const { return myColor; }

  //! Return color.
  const Quantity_Color& Color() const { return myColor.GetRGB(); }

  //! Modifies the color.
  void SetColor (const Quantity_Color& theColor) { myColor.SetRGB (theColor); }

  //! Return line type.
  Aspect_TypeOfLine Type() const { return myType; }

  //! Modifies the type of line.
  void SetType (const Aspect_TypeOfLine theType)
  {
    myType = theType;
    myLinePattern = DefaultLinePatternForType (theType);
  }

  //! Return custom stipple line pattern; 0xFFFF by default.
  uint16_t LinePattern() const { return myLinePattern; }

  //! Modifies the stipple line pattern, and changes line type to Aspect_TOL_USERDEFINED for non-standard pattern.
  void SetLinePattern (uint16_t thePattern)
  {
    myType = DefaultLineTypeForPattern (thePattern);
    myLinePattern = thePattern;
  }

  //! Return stipple line pattern for line type.
  static uint16_t DefaultLinePatternForType (Aspect_TypeOfLine theType)
  {
    switch (theType)
    {
      case Aspect_TOL_DASH:        return 0xFFC0;
      case Aspect_TOL_DOT:         return 0xCCCC;
      case Aspect_TOL_DOTDASH:     return 0xFF18;
      case Aspect_TOL_EMPTY:       return 0x0000;
      case Aspect_TOL_SOLID:       return 0xFFFF;
      case Aspect_TOL_USERDEFINED: return 0xFF24;
    }
    return 0xFFFF;
  }

  //! Return line type for stipple line pattern.
  static Aspect_TypeOfLine DefaultLineTypeForPattern (uint16_t thePattern)
  {
    switch (thePattern)
    {
      case 0x0000: return Aspect_TOL_EMPTY;
      case 0xFFC0: return Aspect_TOL_DASH;
      case 0xCCCC: return Aspect_TOL_DOT;
      case 0xFF18: return Aspect_TOL_DOTDASH;
      case 0xFFFF: return Aspect_TOL_SOLID;
      case 0xFF24: return Aspect_TOL_USERDEFINED;
    }
    return Aspect_TOL_USERDEFINED;
  }

  //! Return a multiplier for each bit in the line stipple pattern within [1, 256] range; 1 by default.
  uint16_t LineStippleFactor() const { return myLineFactor; }

  //! Set a multiplier for each bit in the line stipple pattern.
  void SetLineStippleFactor (uint16_t theFactor)
  {
    if (theFactor == 0 || theFactor > 256)
    {
      throw Standard_OutOfRange ("Graphic3d_Aspects::SetLineStippleFactor(), bad factor value");
    }
    myLineFactor = theFactor;
  }

  //! Return line width.
  Standard_ShortReal Width() const { return myWidth; }

  //! Modifies the line thickness.
  //! Warning: Raises AspectLineDefinitionError if the width is a negative value.
  void SetWidth (const Standard_Real theWidth) { SetWidth ((float )theWidth); }

  //! Modifies the line thickness.
  //! Warning: Raises AspectLineDefinitionError if the width is a negative value.
  void SetWidth (const Standard_ShortReal theWidth)
  {
    if (theWidth <= 0.0f)
    {
      throw Aspect_AspectLineDefinitionError("Graphic3d_AspectLine3d, Bad value for LineWidth");
    }
    myWidth = theWidth;
  }

  //! Return shader program.
  const Handle(Graphic3d_ShaderProgram)& ShaderProgram() const { return myProgram; }

  //! Sets up OpenGL/GLSL shader program.
  void SetShaderProgram (const Handle(Graphic3d_ShaderProgram)& theProgram) { myProgram = theProgram; }

  //! Check for equality with another line aspect.
  bool IsEqual (const Graphic3d_AspectLine3d& theOther)
  {
    if (this == &theOther)
    {
      return true;
    }

    return myProgram == theOther.myProgram
        && myType    == theOther.myType
        && myLineFactor == theOther.myLineFactor
		&& myLinePattern == theOther.myLinePattern
        && myColor   == theOther.myColor
        && myWidth   == theOther.myWidth;
  }

public:

  //! Returns the current values of the group.
  Standard_DEPRECATED("Deprecated method Values() should be replaced by individual property getters")
  void Values (Quantity_Color&    theColor,
               Aspect_TypeOfLine& theType,
               Standard_Real&     theWidth) const
  {
    theColor = myColor.GetRGB();
    theType  = myType;
    theWidth = myWidth;
  }

protected:

  Handle(Graphic3d_ShaderProgram) myProgram;
  Quantity_ColorRGBA myColor;
  Aspect_TypeOfLine  myType;
  Standard_ShortReal myWidth;
  uint16_t           myLineFactor;
  uint16_t           myLinePattern;

};

DEFINE_STANDARD_HANDLE(Graphic3d_AspectLine3d, Standard_Transient)

#endif // _Graphic3d_AspectLine3d_HeaderFile
